# -*- coding: utf-8 -*-
"default plugin for setting: set it in a simple dictionary"
# Copyright (C) 2013-2019 Team tiramisu (see AUTHORS for all contributors)
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ____________________________________________________________
from .sqlite3db import Sqlite3DB
from ...log import log


class Properties(Sqlite3DB):
    __slots__ = tuple()

    def __init__(self, storage):
        super(Properties, self).__init__(storage)

    # properties
    def setproperties(self, path, index, properties):
        self.exportation()
        self.delproperties(path, index, commit=False)
        self.exportation()
        self._storage.execute("INSERT INTO property(path, tiram_index, properties, session_id) VALUES "
                              "(?, ?, ?, ?)", (path,
                                               index,
                                               self._sqlite_encode(properties),
                                               self._session_id))
        self.exportation()

    def getproperties(self, path, index, default_properties):
        sql = 'SELECT properties FROM property WHERE session_id = ? '
        params = [self._session_id]
        if path is None:
             sql += "AND path is NULL "
        else:
             sql += "AND path = ? "
             params.append(path)
        if index is None:
             sql += "AND tiram_index is NULL LIMIT 1"
        else:
             sql += "AND tiram_index = ? LIMIT 1"
             params.append(index)
        value = self._storage.select(sql, params)
        if value is None:
            return set(default_properties)
        else:
            return set(self._sqlite_decode(value[0]))

    def delproperties(self, path, index, commit=True):
        sql = 'DELETE FROM property WHERE session_id = ? '
        params = [self._session_id]
        if path is None:
            sql += 'AND path is NULL '
        else:
            sql += 'AND path = ? '
            params.append(path)
        if index is None:
            sql += 'AND tiram_index is NULL'
        else:
            params.append(index)
            sql += 'AND tiram_index = ?'
        self._storage.execute(sql, params, commit)

    def exportation(self):
        """return all modified settings in a dictionary
        example: {'path1': set(['prop1', 'prop2'])}
        """
        ret = {}
        for path, tiram_index, properties, _ in self._storage.select("SELECT * FROM property "
                                                                     "WHERE session_id = ?",
                                                                     (self._session_id,),
                                                                     only_one=False):
            ret.setdefault(path, {})[tiram_index] = self._sqlite_decode(properties)
        return ret

    def importation(self, properties):
        self._storage.execute("DELETE FROM property WHERE session_id = ?", (self._session_id,), commit=False)
        for path, indexed_properties in properties.items():
            for index, property_ in indexed_properties.items():
                self._storage.execute("INSERT INTO property(path, tiram_index, properties, session_id) "
                                      "VALUES (?, ?, ?, ?)", (path,
                                                              index,
                                                              self._sqlite_encode(property_),
                                                              self._session_id,
                                                              ), False)
        self._storage._conn.commit()


class Permissives(Sqlite3DB):
    __slots__ = tuple()

    # permissive
    def setpermissives(self, path, index, permissive):
        log.debug('setpermissive %s %s %s %s', path, index, permissive, id(self))
        self.delpermissive(path, index, commit=False)
        self._storage.execute("INSERT INTO permissive(path, tiram_index, permissives, session_id) "
                              "VALUES (?, ?, ?, ?)", (path,
                                                      index,
                                                      self._sqlite_encode(permissive),
                                                      self._session_id))

    def getpermissives(self, path, index):
        sql = 'SELECT permissives FROM permissive WHERE session_id = ? '
        params = [self._session_id]
        if path is None:
             sql += "AND path is NULL "
        else:
             sql += "AND path = ? "
             params.append(path)
        if index is None:
             sql += "AND tiram_index is NULL LIMIT 1"
        else:
             sql += "AND tiram_index = ? LIMIT 1"
             params.append(index)
        permissives = self._storage.select(sql, params)
        if permissives is None:
            ret = frozenset()
        else:
            ret = frozenset(self._sqlite_decode(permissives[0]))
        log.debug('getpermissive %s %s %s', path, ret, id(self))
        return ret

    def delpermissive(self, path, index, commit=True):
        sql = 'DELETE FROM permissive WHERE session_id = ? '
        params = [self._session_id]
        if path is None:
            sql += 'AND path is NULL '
        else:
            sql += 'AND path = ? '
            params.append(path)
        if index is None:
            sql += 'AND tiram_index is NULL'
        else:
            params.append(index)
            sql += 'AND tiram_index = ?'
        self._storage.execute(sql, params, commit)

    def exportation(self):
        """return all modified permissives in a dictionary
        example: {'path1': set(['perm1', 'perm2'])}
        """
        ret = {}
        for path, index, permissives in self._storage.select("SELECT path, tiram_index, permissives FROM permissive "
                                                      "WHERE session_id = ?",
                                                      (self._session_id,),
                                                      only_one=False):
            ret.setdefault(path, {})[index] = self._sqlite_decode(permissives)
        return ret

    def importation(self, permissives):
        self._storage.execute("DELETE FROM permissive WHERE session_id = ?", (self._session_id,),
                              commit=False)
        for path, indexed_permissives in permissives.items():
            for index, permissive in indexed_permissives.items():
                self._storage.execute("INSERT INTO permissive(path, tiram_index, permissives, session_id) "
                                      "VALUES (?, ?, ?, ?)", (path,
                                                              index,
                                                              self._sqlite_encode(permissive),
                                                              self._session_id,
                                                              ), False)
        self._storage._conn.commit()
